// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

pub mod arrow_parquet;
pub mod base_writer;
pub mod config;
pub mod file_writer;
pub mod partition;
pub mod table_writer;

use crate::spec::DataFile;

pub trait IcebergWriter<T> {
    fn add(&mut self, data: T) -> Result<(), String>;
    fn close(self) -> Result<Vec<DataFile>, String>;
}

pub struct WriteOutcome {
    pub data_file: DataFile,
}
