/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

syntax = 'proto3';

import "spark/connect/relations.proto";
import "spark/connect/expressions.proto";
option go_package = "internal/generated";

package spark.connect;

option java_multiple_files = true;
option java_package = "org.apache.spark.connect.proto";

message ExamplePluginRelation {
  Relation input = 1;
  string custom_field = 2;

}

message ExamplePluginExpression {
  Expression child = 1;
  string custom_field = 2;
}

message ExamplePluginCommand {
  string custom_field = 1;
}
